# -*- coding: utf-8 -*-

from collections import OrderedDict
import datetime
import json

from gisdo import constants as const
from gisdo.alchemy_session import session_scope
from gisdo.algorithm.enrolled import get_22_1_index
from gisdo.algorithm.enrolled import get_22_2_index
from gisdo.algorithm.enrolled import get_22_3_index
from gisdo.algorithm.enrolled import get_22_4_index
from gisdo.algorithm.enrolled import get_22_5_index
from gisdo.algorithm.enrolled import get_22_6_index
from gisdo.algorithm.enrolled import get_22_7_index
from gisdo.algorithm.enrolled import get_22_8_index
from gisdo.algorithm.enrolled import get_22_index
from gisdo.algorithm.enrolled import get_30_index
from gisdo.algorithm.enrolled import get_30_1_index
from gisdo.algorithm.enrolled import get_30_2_index
from gisdo.algorithm.direct import get_direct_ids
from gisdo.algorithm.direct import get_32_filter
from gisdo.algorithm.direct import get_321_filter
from gisdo.algorithm.direct import get_322_filter
from gisdo.exceptions import NoDataInMo
from gisdo.index import ApplicationIndex
from gisdo.index import DirectIndex
from gisdo.index import get_enrolled_age_filter
from gisdo.index import get_enrolled_count
from gisdo.index import get_queue_age_filter
from gisdo.index import get_queue_average_waiting_time
from gisdo.index import get_queue_count
from gisdo.index import get_queue_index
from gisdo.index import get_queue_index_collection
from gisdo.index.capacities import free_space_in_age_category
from gisdo.index.constants import get_all_day_work_types
from gisdo.index.constants import get_short_day_types
from gisdo.index.direct import ind_19_3
from gisdo.index.direct import ind_29_1
from gisdo.index.enrolled import Index
from gisdo.models import ReportFormRow
from gisdo.utils import AverageWaitingTimeWrapperEncoder
from gisdo.utils import DateMixin
from kinder.core.dict.models import GroupTypeEnumerate


class ReportCacheMixin(object):

    APPLICATIONS = 'applications'
    QUEUES = 'queues'
    ENROLLED = 'enrolled'
    CAPACITIES = 'capacities'

    DATA_TYPE = [
        APPLICATIONS,
        QUEUES,
        ENROLLED,
        CAPACITIES
    ]

    def __init__(self):

        self._cache = {
            self.APPLICATIONS: {},
            self.QUEUES: {},
            self.ENROLLED: {},
            self.CAPACITIES: {}
        }

    def _update_cache(self, data, data_type):
        u"""Обновление кэша."""

        for key, value in data.iteritems():
            if key not in self._cache[data_type]:
                self._cache[data_type][key] = value
            elif isinstance(value, dict):
                for age_cat in data[key]:
                    self._cache[
                        data_type][key][age_cat] += data[key][age_cat]


class DataCollector(ReportCacheMixin):
    def __init__(self, mo, unit_helper):

        super(DataCollector, self).__init__()

        self._mo = mo
        self._unit_helper = unit_helper

    def collect(self, report_form):

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        # Для расчета показателя 32 и 32.1, 32.2, необходим список
        #  идентификаторов последних направлений у ребенка в рамках МО
        self.list_last_id_directs_32 = get_direct_ids(
            self._mo.id, get_32_filter)
        self.list_last_id_directs_321 = get_direct_ids(
            self._mo.id, get_321_filter)
        self.list_last_id_directs_322 = get_direct_ids(
            self._mo.id, get_322_filter)

        # Собираем по ДОУ
        for dou in dou_list:
            self._dou_collect(report_form, dou)

        # Данные по МО лежат в кэше
        ReportFormRow.objects.get_or_create(
            applications=json.dumps(
                self._cache[self.APPLICATIONS]),
            queues=json.dumps(
                self._cache[self.QUEUES], cls=AverageWaitingTimeWrapperEncoder),
            enrolled=json.dumps(self._cache[self.ENROLLED]),
            capacities=json.dumps(
                self._cache[self.CAPACITIES]),
            report=report_form,
            unit=self._mo)

        return self._cache

    def _dou_collect(self, report_form, dou):

        applications = self._collect_applications_data(dou)
        queues = self._collect_queue_data(dou)
        enrolled = self._collect_enrolled_data(dou)
        capacities = self._collect_capacities(dou)

        ReportFormRow.objects.get_or_create(
            report=report_form,
            unit=dou,
            applications=json.dumps(applications),
            queues=json.dumps(queues, cls=AverageWaitingTimeWrapperEncoder),
            enrolled=json.dumps(enrolled),
            capacities=json.dumps(capacities),
        )

    @staticmethod
    def _merge_dicts(first_dict, second_dict):
        for index in first_dict:
            for age_cat in first_dict[index]:
                try:
                    value = second_dict[index][age_cat]
                except KeyError:
                    value = 0
                first_dict[index][age_cat] += value

    def _collect_applications_data(self, dou):

        def _inner(dou):

            applications = OrderedDict([
                ('1', OrderedDict()),
                ('1.1', OrderedDict()),
                ('2.1', OrderedDict()),
                ('3.1', OrderedDict()),
                ('4', OrderedDict()),
                ('4.1', OrderedDict()),
                ('4.2', OrderedDict()),
                ('5', OrderedDict()),
                ('6', OrderedDict()),
            ])

            index_4 = get_queue_index_collection("4", dou)
            index_4_1 = get_queue_index_collection("4.1", dou)
            index_4_2 = get_queue_index_collection("4.2", dou)

            index_5 = get_queue_index_collection("5", dou)

            with session_scope() as session:
                index = ApplicationIndex(dou, session)

                for value in const.AGE_CATEGORIES_FULL.values():

                    on_date_filter = get_queue_age_filter(
                        value, datetime.date.today
                    )

                    applications['2.1'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                        portal=1,
                        age_range=value,
                        on_date=datetime.date.today,
                    )

                    applications['3.1'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_DELIVERY,
                        age_range=value,
                        on_date=datetime.date.today,
                    )

                    applications['1.1'][value] = (
                        applications['2.1'][value] + applications['3.1'][value]
                    )

                    applications["4"][value] = get_queue_count(
                        index_4, on_date_filter)
                    applications["4.1"][value] = get_queue_count(
                        index_4_1, on_date_filter)
                    applications["4.2"][value] = get_queue_count(
                        index_4_2, on_date_filter)

                    applications["5"][value] = get_queue_count(
                        index_5, on_date_filter)

                    # Общее количество заявлений на постановку в очередь
                    # в группы кратковременного пребывания
                    applications['6'][value] = index.get_count(
                        index_type=ApplicationIndex.COUNT_BY_GROUP_TYPE,
                        age_range=value,
                        on_date=datetime.date.today,
                    )

                applications['1'][const.ALL] = get_queue_index('1', dou)

            return applications

        app = _inner(dou)
        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_app = _inner(filial)
            self._merge_dicts(app, filial_app)

        self._update_cache(app, self.APPLICATIONS)

        return app

    def _collect_queue_data(self, dou):

        def _inner(dou):
            queues = OrderedDict([
                ('7', OrderedDict()), ('7.1', OrderedDict()),
                ('7.2', OrderedDict()), ('7.3', OrderedDict()),
                ('7.4', OrderedDict()), ('7.5', OrderedDict()),
                ('7.6', OrderedDict()), ('7.7', OrderedDict()),
                ('8', OrderedDict()),
                ('8.1', OrderedDict()), ('8.2', OrderedDict()),
                ('8.3', OrderedDict()),
                ('9', OrderedDict()), ('9.1', OrderedDict()),
                ('10', OrderedDict()), ('10.1', OrderedDict()),
                ('11', OrderedDict()),
                ('12', OrderedDict()), ('12.1', OrderedDict()),
                ('13', OrderedDict()),
                ('13.1', OrderedDict()), ('14', OrderedDict()),
                ('15', OrderedDict()), ('16', OrderedDict()),
                ('17', OrderedDict()), ('17.1', OrderedDict()),
                ('18', OrderedDict()), ('18.1', OrderedDict()),
                ('18.2', OrderedDict()), ('18.3', OrderedDict()),
                ('18.4', OrderedDict()), ('18.5', OrderedDict()),
            ])

            index_7 = get_queue_index_collection("7", dou)
            index_7_1 = get_queue_index_collection("7.1", dou)
            index_7_2 = get_queue_index_collection("7.2", dou)
            index_7_3 = get_queue_index_collection("7.3", dou)
            index_7_4 = get_queue_index_collection('7.4', dou)
            index_7_5 = get_queue_index_collection('7.5', dou)
            index_7_7 = get_queue_index_collection('7.7', dou)

            index_8 = get_queue_index_collection('8', dou)
            index_8_1 = get_queue_index_collection('8.1', dou)
            index_8_2 = get_queue_index_collection('8.2', dou)
            index_8_3 = get_queue_index_collection('8.3', dou)

            index_10 = get_queue_index_collection("10", dou)
            index_10_1 = get_queue_index_collection("10.1", dou)

            index_11 = get_queue_index_collection("11", dou)

            index_12 = get_queue_index_collection("12", dou)
            index_12_1 = get_queue_index_collection("12.1", dou)

            index_13 = get_queue_index_collection("13", dou)
            index_13_1 = get_queue_index_collection("13.1", dou)

            index_14 = get_queue_index_collection("14", dou)

            index_15 = get_queue_index_collection("15", dou)

            index_16 = get_queue_index_collection("16", dou)

            index_17 = get_queue_index_collection("17", dou)
            index_17_1 = get_queue_index_collection("17.1", dou)

            index_18 = get_queue_index_collection("18", dou)
            index_18_1 = get_queue_index_collection('18.1', dou)
            index_18_2 = get_queue_index_collection("18.2", dou)
            index_18_3 = get_queue_index_collection('18.3', dou)
            index_18_4 = get_queue_index_collection('18.4', dou)
            index_18_5 = get_queue_index_collection('18.5', dou)

            queues['10'][const.ALL] = 0
            queues['11'][const.ALL] = 0

            for value in const.AGE_CATEGORIES_FULL.values():

                on_date_filter = get_queue_age_filter(
                    value, datetime.date.today
                )
                current_year_first_september = get_queue_age_filter(
                    value,
                    DateMixin.get_current_calendar_year(),
                )
                next_year_first_day = get_queue_age_filter(
                    value,
                    datetime.date(datetime.date.today().year+1, 1, 1),
                )
                next_learn_year_first_day = get_queue_age_filter(
                    value,
                    DateMixin._get_next_calendar_year(),
                )

                # Отложенный спрос
                queues["7"][value] = get_queue_count(
                    index_7, current_year_first_september)
                queues["7.1"][value] = get_queue_count(
                    index_7_1, on_date_filter)
                queues["7.2"][value] = get_queue_count(
                    index_7_2, on_date_filter)
                queues["7.4"][value] = get_queue_count(
                    index_7_4, on_date_filter)
                queues["7.5"][value] = get_queue_count(
                    index_7_5, on_date_filter)
                queues["7.6"][value] = get_queue_count(
                    index_7_5, next_learn_year_first_day)
                queues["7.7"][value] = get_queue_count(
                    index_7_7, next_learn_year_first_day)

                # Актуальный спрос
                queues["7.3"][value] = get_queue_count(
                    index_7_3, on_date_filter)
                queues["18"][value] = get_queue_count(
                    index_18, current_year_first_september)
                queues["18.1"][value] = get_queue_count(
                    index_18_1, on_date_filter)
                queues["18.2"][value] = get_queue_count(
                    index_18_2, next_year_first_day)
                queues["18.3"][value] = get_queue_count(
                    index_18_3, on_date_filter)
                queues["18.4"][value] = get_queue_count(
                    index_18_4, on_date_filter)
                queues["18.5"][value] = get_queue_count(
                    index_18_5, on_date_filter)

                # Количество детей в очереди с ограниченными
                # возможностями по здоровью
                queues["12"][value] = get_queue_count(index_12, on_date_filter)
                queues["12.1"][value] = get_queue_count(
                    index_12_1, on_date_filter)

                queues["13"][value] = get_queue_count(index_13, on_date_filter)
                queues["13.1"][value] = get_queue_count(
                    index_13_1, on_date_filter)

                queues["14"][value] = get_queue_count(
                    index_14, on_date_filter)

                queues["15"][value] = get_queue_count(index_15, on_date_filter)

                queues["16"][value] = get_queue_count(index_16, on_date_filter)

                queues["17"][value] = get_queue_count(index_17, on_date_filter)
                queues["17.1"][value] = get_queue_count(
                    index_17_1, on_date_filter)

                queues["9"][value] = get_queue_count(
                    index_10, current_year_first_september
                ) + get_queue_count(index_11, current_year_first_september)

                queues["10.1"][value] = get_queue_count(
                    index_10_1, on_date_filter)

                # Количество детей в очереди имеющих право зачисления
                # на основе федеральных и региональных льгот.
                current_10_value = get_queue_count(index_10, on_date_filter)
                queues['10'][const.ALL] += current_10_value

                current_11_value = get_queue_count(index_11, on_date_filter)
                queues['11'][const.ALL] += current_11_value

                queues['9.1'][value] = current_10_value + current_11_value

            for value in const.AGE_CATEGORIES_EIGHT.values():

                queues["8"][value] = get_queue_average_waiting_time(
                    index_8, value)
                queues["8.1"][value] = get_queue_average_waiting_time(
                    index_8_1, value)
                queues["8.2"][value] = get_queue_average_waiting_time(
                    index_8_2, value)
                queues["8.3"][value] = get_queue_average_waiting_time(
                    index_8_3, value)

            return queues

        queues = _inner(dou)
        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_queues = _inner(filial)
            self._merge_dicts(queues, filial_queues)

        self._update_cache(queues, self.QUEUES)

        return queues

    def _collect_enrolled_data(self, dou):

        def _inner(dou, calc_filial=False):
            enrolled = OrderedDict([
                ('19', OrderedDict()), ('19.1', OrderedDict()),
                ('19.2', OrderedDict()), ('19.3', OrderedDict()),
                ('20', OrderedDict()), ('20.1', OrderedDict()),
                ('20.2', OrderedDict()), ('20.3', OrderedDict()),
                ('20.4', OrderedDict()), ('20.5', OrderedDict()),
                ('20.6', OrderedDict()), ('20.7', OrderedDict()),
                ('20.8', OrderedDict()),
                ('21', OrderedDict()),
                ('22', OrderedDict()), ('22.1', OrderedDict()),
                ('22.2', OrderedDict()), ('22.3', OrderedDict()),
                ('22.4', OrderedDict()), ('22.5', OrderedDict()),
                ('22.6', OrderedDict()), ('22.7', OrderedDict()),
                ('22.8', OrderedDict()), ('23', OrderedDict()),
                ('24', OrderedDict()), ('25', OrderedDict()),
                ('26', OrderedDict()), ('27', OrderedDict()),
                ('28', OrderedDict()), ('29', OrderedDict()),
                ('29.1', OrderedDict()), ('30', OrderedDict()),
                ('30.1', OrderedDict()), ('30.2', OrderedDict()),
            ])

            enrolled_index = Index(dou)

            index_20 = get_queue_index_collection("20", dou)
            index_20_1 = get_queue_index_collection(
                "20.1", dou, check_unit=False)
            index_20_2 = get_queue_index_collection("20.2", dou)
            index_20_3 = get_queue_index_collection("20.3", dou)
            index_20_5 = get_queue_index_collection(
                "20.5", dou, check_unit=False)
            index_20_6 = get_queue_index_collection(
                "20.6", dou, check_unit=False)
            index_20_7 = get_queue_index_collection(
                "20.7", dou, check_unit=False)
            index_20_8 = get_queue_index_collection(
                "20.8", dou, check_unit=False)
            index_29_1 = ind_29_1(dou)
            if not calc_filial:
                index_19 = list(enrolled_index(force=True))
                index_21 = list(enrolled_index(
                    force=True, group_filter=lambda group: (
                        group.type and
                        group.type.code == GroupTypeEnumerate.FAMILY
                    )))
                index_22 = get_22_index(enrolled_index)
                index_22_1 = get_22_1_index(enrolled_index)
                index_22_2 = get_22_2_index(enrolled_index)
                index_22_3 = get_22_3_index(enrolled_index)
                index_22_4 = get_22_4_index(enrolled_index)
                index_22_5 = get_22_5_index(enrolled_index)
                index_22_6 = get_22_6_index(enrolled_index)
                index_22_7 = get_22_7_index(enrolled_index)
                index_22_8 = get_22_8_index(enrolled_index)

                index_23 = list(
                    enrolled_index(force=True, group_filter=lambda group: (
                        group.type and
                        group.type.code == GroupTypeEnumerate.COMP
                    ))
                )

                index_24 = list(
                    enrolled_index(force=True, group_filter=lambda group: (
                        group.type and
                        group.type.code == GroupTypeEnumerate.HEALTH
                    ))
                )

                index_25 = list(
                    enrolled_index(force=True, group_filter=lambda group: (
                        group.type and
                        (group.type.code == GroupTypeEnumerate.CARE or
                         group.type.code == GroupTypeEnumerate.YOUNG)
                    ))
                )

                index_26 = list(
                    enrolled_index(force=True, group_filter=lambda group: (
                        group.work_type and
                        group.work_type.code in get_all_day_work_types()
                    ))
                )

                index_27 = list(
                    enrolled_index(force=True, group_filter=lambda group: (
                        group.work_type and
                        group.work_type.code in get_short_day_types()
                    ))
                )

                # дети из групп кратковременного пребывания но
                # посещающие ДОО в другом режиме
                short_wt_groups_pupils_with_other_wt_idx = list(
                    enrolled_index(
                        force=True,
                        group_filter=lambda group: (
                                group.work_type and
                                group.work_type.code in get_short_day_types()
                        ),
                        query_filter='not_short_actual_wt',
                    )
                )

                # дети не из групп кратковременного пребывания но
                # посещающие ДОО с этим режимом
                not_short_wt_groups_pupils_with_short_status_idx = list(
                    enrolled_index(
                        force=True,
                        group_filter=lambda group: not (
                                group.work_type and
                                group.work_type.code in get_short_day_types()
                        ),
                        query_filter='short_actual_wt',
                    ),
                )

                index_30 = get_30_index(enrolled_index)
                index_30_1 = get_30_1_index(enrolled_index)
                index_30_2 = get_30_2_index(enrolled_index)

            for value in const.AGE_CATEGORIES_FULL.values():
                current_year_enrolled_filter = get_enrolled_age_filter(
                    value, DateMixin.get_current_calendar_year())
                on_date_enrolled_filter = get_enrolled_age_filter(
                    value, datetime.date.today)
                next_calendar_year_enrolled = get_enrolled_age_filter(
                    value, DateMixin.get_next_calendar_year)
                # Общая численность детей зачисленных в ДОО
                if not calc_filial:
                    enrolled["19"][value] = get_enrolled_count(
                        index_19, current_year_enrolled_filter)
                    enrolled["19.1"][value] = get_enrolled_count(
                        index_19, on_date_enrolled_filter)
                    enrolled["19.2"][value] = get_enrolled_count(
                        index_19, next_calendar_year_enrolled)
                else:
                    enrolled["19"][value] = 0
                    enrolled["19.1"][value] = 0
                    enrolled["19.2"][value] = 0

                enrolled['19.3'][value] = ind_19_3(dou, age_range=value)

                on_date_filter = get_queue_age_filter(
                    value, datetime.date.today)
                enrolled["20"][value] = get_queue_count(
                    index_20, on_date_filter)
                enrolled["20.1"][value] = get_queue_count(
                    index_20_1, on_date_filter)
                enrolled["20.2"][value] = get_queue_count(
                    index_20_2, on_date_filter)
                enrolled["20.3"][value] = get_queue_count(
                    index_20_3, on_date_filter)
                enrolled["20.5"][value] = get_queue_count(
                    index_20_5, on_date_filter)
                enrolled["20.6"][value] = get_queue_count(
                    index_20_6, on_date_filter)
                enrolled["20.7"][value] = get_queue_count(
                    index_20_7, on_date_filter)
                enrolled["20.8"][value] = get_queue_count(
                    index_20_8, on_date_filter)

                # Общая численность детей, зачисленных в группы для
                # детей с ограниченными возможностями здоровью
                if not calc_filial:
                    enrolled["21"][value] = get_enrolled_count(
                        index_21, on_date_enrolled_filter)
                    enrolled["22"][value] = get_enrolled_count(
                        index_22, on_date_enrolled_filter)
                    # Общая численность детей, зачисленных в группы для
                    # детей с ограниченными возможностями здоровью
                    enrolled["22"][value] = get_enrolled_count(
                        index_22, on_date_enrolled_filter)
                    enrolled["22.1"][value] = get_enrolled_count(
                        index_22_1, on_date_enrolled_filter)
                    enrolled["22.2"][value] = get_enrolled_count(
                        index_22_2, on_date_enrolled_filter)
                    enrolled["22.3"][value] = get_enrolled_count(
                        index_22_3, on_date_enrolled_filter)
                    enrolled["22.4"][value] = get_enrolled_count(
                        index_22_4, on_date_enrolled_filter)
                    enrolled["22.5"][value] = get_enrolled_count(
                        index_22_5, on_date_enrolled_filter)
                    enrolled["22.6"][value] = get_enrolled_count(
                        index_22_6, on_date_enrolled_filter)
                    enrolled["22.7"][value] = get_enrolled_count(
                        index_22_7, on_date_enrolled_filter)
                    enrolled["22.8"][value] = get_enrolled_count(
                        index_22_8, on_date_enrolled_filter)
                else:
                    enrolled['21'][value] = enrolled["22"][value] = 0
                    enrolled['22.1'][value] = enrolled['22.2'][value] = 0
                    enrolled['22.3'][value] = enrolled['22.4'][value] = 0
                    enrolled['22.5'][value] = enrolled['22.6'][value] = 0
                    enrolled['22.7'][value] = enrolled['22.8'][value] = 0
                # Общее количество детей зачисленных в
                # оздоровительные и компенсирующие группы
                if not calc_filial:
                    enrolled["23"][value] = get_enrolled_count(
                        index_23, on_date_enrolled_filter)
                    enrolled["24"][value] = get_enrolled_count(
                        index_24, on_date_enrolled_filter)
                    enrolled["25"][value] = get_enrolled_count(
                        index_25, on_date_enrolled_filter)
                    # Общая численность детей, зачисленных в
                    # группы круглосуточного пребывания детей
                    enrolled["26"][value] = get_enrolled_count(
                        index_26, on_date_enrolled_filter)
                    # Общая численность детей, зачисленных в группы
                    # кратковременного пребывания детей
                    enrolled["27"][value] = get_enrolled_count(
                        index_27, on_date_enrolled_filter)
                    # Общая численность детей, посещающих ДОО в режиме
                    # кратковременного пребывания
                    enrolled['28'][value] = (
                        enrolled['27'][value]
                        - get_enrolled_count(
                            short_wt_groups_pupils_with_other_wt_idx,
                            on_date_enrolled_filter
                        )
                        + get_enrolled_count(
                            not_short_wt_groups_pupils_with_short_status_idx,
                            on_date_enrolled_filter
                        )
                    )
                else:
                    enrolled["23"][value] = enrolled["24"][value] = 0
                    enrolled["25"][value] = enrolled["26"][value] = 0
                    enrolled["27"][value] = enrolled['28'][value] = 0

                if not calc_filial:
                    enrolled['30'][value] = get_enrolled_count(
                        index_30, on_date_enrolled_filter)
                    enrolled['30.1'][value] = get_enrolled_count(
                        index_30_1, on_date_enrolled_filter)
                    enrolled['30.2'][value] = get_enrolled_count(
                        index_30_2, on_date_enrolled_filter)
                else:
                    enrolled['30'][value] = enrolled['30.1'][value] = 0
                    enrolled['30.2'][value] = 0

                enrolled['29.1'][value] = get_enrolled_count(
                    index_29_1, on_date_enrolled_filter)

                enrolled['20.4'][value] = get_queue_index(
                    "20.4", dou, age_range=value)

            # Общее количество детей зачисленных в семейные,
            # оздоровительные и компенсирующие группы
            if not calc_filial:
                enrolled['29'][const.ALL] = len(list(
                    enrolled_index(
                        force=True,
                        query_filter='predictable_decrease_contingent')))
            else:
                enrolled['29'][const.ALL] = 0

            return enrolled

        enrolled = _inner(dou)
        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_enrolled = _inner(filial, calc_filial=True)
            self._merge_dicts(enrolled, filial_enrolled)

        self._update_cache(enrolled, self.ENROLLED)

        return enrolled

    def _collect_capacities(self, dou):

        def _inner(dou, calc_filial=True):
            capacities = OrderedDict([
                ('31', OrderedDict()), ('31.1', OrderedDict()),
                ('31.2', OrderedDict()), ('31.3', OrderedDict()),
                ('31.4', OrderedDict()), ('32', OrderedDict()),
                ('32.1', OrderedDict()),('32.2', OrderedDict()),
                ('33', OrderedDict())
            ])

            if calc_filial:
                group_index = Index(dou)
                if not group_index.is_calculated:
                    common = []
                    short_stay = []
                    with_hn = []
                    compensating = []
                    health_group = []
                else:
                    common = group_index.free_space_in_unit(
                        force=True, group_filter=None)
                    short_stay = group_index.free_space_in_unit(
                        force=False, group_filter='short_stay')
                    with_hn = group_index.free_space_in_unit(
                        force=False, group_filter='with_health_need')
                    compensating = group_index.free_space_in_unit(
                        force=False, group_filter='compensating')
                    health_group = group_index.free_space_in_unit(
                        force=False, group_filter='health_group')

            direct_index = DirectIndex(dou)

            for value in const.AGE_CATEGORIES_EIGHT.values():
                if calc_filial:
                    capacities['31'][value] = free_space_in_age_category(
                        common, value)
                    capacities['31.1'][value] = free_space_in_age_category(
                        with_hn, value)
                    capacities['31.2'][value] = free_space_in_age_category(
                        compensating, value)
                    capacities['31.3'][value] = free_space_in_age_category(
                        health_group, value)
                    capacities['31.4'][value] = free_space_in_age_category(
                        short_stay, value)
                else:
                    capacities['31'][value] = capacities['31.1'][value] = 0
                    capacities['31.2'][value] = capacities['31.3'][value] = 0
                    capacities['31.4'][value] = 0

            capacities['32'][const.ALL] = direct_index.get_count(
                self.list_last_id_directs_32)
            for value in const.AGE_CATEGORIES_FULL.values():
                capacities['32.1'][value] = direct_index.get_count(
                    self.list_last_id_directs_321,
                    age_range=value,
                    on_date=datetime.date.today
                )
                capacities['32.2'][value] = direct_index.get_count(
                    self.list_last_id_directs_322,
                    age_range=value,
                    on_date=datetime.date.today
                )

            capacities['33'][const.ALL] = get_queue_index("33", dou)

            return capacities

        capacities = _inner(dou)
        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_capacities = _inner(filial, calc_filial=False)
            self._merge_dicts(capacities, filial_capacities)

        self._update_cache(capacities, self.CAPACITIES)

        return capacities
